/*
 * conf.c
 *
 * This file is part of msmtp, an SMTP client.
 *
 * Copyright (C) 2000, 2003, 2004, 2005
 * Martin Lambers <marlam@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *   msmtp is released under the GPL with the additional exemption that
 *   compiling, linking, and/or using OpenSSL is allowed.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
extern int errno;
#ifdef WINDOWS
#include "winsock2.h"
#else
#include <netdb.h>
#include <arpa/inet.h>
#endif

#include "paths.h"
#include "list.h"
#include "xmalloc.h"
#include "conf.h"

/* buffer size for configuration file lines */
#define LINEBUFSIZE 501


/* 
 * account_new()
 * 
 * see conf.h
 */

account_t *account_new(const char *conffile, const char *id)
{
    account_t *a;
    a = xmalloc(sizeof(account_t));
    a->id = id ? xstrdup(id) : NULL;
    a->conffile = conffile ? xstrdup(conffile) : NULL;
    a->mask = 0;
    a->host = NULL;
    a->port = 0;
    a->domain = xstrdup("localhost");
    a->from = NULL;
    a->dsn_return = NULL;
    a->dsn_notify = NULL;
    a->keepbcc = 0;
    a->auth_mech = NULL;
    a->username = NULL;
    a->password = NULL;
    a->ntlmdomain = NULL;
    a->tls = 0;
    a->tls_key_file = NULL;
    a->tls_cert_file = NULL;
    a->tls_trust_file = NULL;
    a->tls_nostarttls = 0;
    a->tls_nocertcheck = 0;
    a->logfile = NULL;
    return a;
}


/*
 * account_copy()
 *
 * see conf.h
 */

account_t *account_copy(account_t *acc)
{
    account_t *a = NULL;
    
    if (acc)
    {
	a = xmalloc(sizeof(account_t));
	a->id = acc->id ? xstrdup(acc->id) : NULL;
	a->conffile = acc->conffile ? xstrdup(acc->conffile) : NULL;
	a->mask = acc->mask;
	a->host = acc->host ? xstrdup(acc->host) : NULL;
	a->port = acc->port;
	a->domain = acc->domain ? xstrdup(acc->domain) : NULL;
	a->from = acc->from ? xstrdup(acc->from) : NULL;
	a->dsn_return = acc->dsn_return ? xstrdup(acc->dsn_return) : NULL;
	a->dsn_notify = acc->dsn_notify ? xstrdup(acc->dsn_notify) : NULL;
	a->keepbcc = acc->keepbcc;
	a->auth_mech = acc->auth_mech ? xstrdup(acc->auth_mech) : NULL;
	a->username = acc->username ? xstrdup(acc->username) : NULL;
	a->password = acc->password ? xstrdup(acc->password) : NULL;
	a->ntlmdomain = acc->ntlmdomain ? xstrdup(acc->ntlmdomain) : NULL;
	a->tls = acc->tls;
	a->tls_key_file = acc->tls_key_file ? xstrdup(acc->tls_key_file) : NULL;
	a->tls_cert_file = acc->tls_cert_file ? xstrdup(acc->tls_cert_file) : NULL;
	a->tls_trust_file = acc->tls_trust_file ? xstrdup(acc->tls_trust_file) : NULL;
	a->tls_nostarttls = acc->tls_nostarttls;
	a->tls_nocertcheck = acc->tls_nocertcheck;
	a->logfile = acc->logfile ? xstrdup(acc->logfile) : NULL;
    }
    return a;
}


/* 
 * account_free()
 *
 * see conf.h
 */

void account_free(void *a)
{
    account_t *p = a;
    if (p)
    {
	free(p->id);
	free(p->conffile);
	free(p->host);
	free(p->domain);
	free(p->from);
	free(p->auth_mech);
	free(p->username);
	free(p->password);
	free(p->ntlmdomain);
	free(p->tls_key_file);
	free(p->tls_cert_file);
	free(p->tls_trust_file);
	free(p->dsn_return);
	free(p->dsn_notify);
	free(p->logfile);
	free(p);
    }
}


/*
 * find_account()
 *
 * see conf.h
 */

account_t *find_account(list_t *acc_list, const char *id)
{
    account_t *a = NULL;
    char *acc_id;
    
    while (!list_is_empty(acc_list))
    {
	acc_list = acc_list->next;
	acc_id = ((account_t *)(acc_list->data))->id;
	if (acc_id && strcmp(id, acc_id) == 0)
	{
	    a = acc_list->data;
	    break;
	}
    }

    return a;
}


/*
 * find_account_by_envelope_from()
 *
 * see conf.h
 */

account_t *find_account_by_envelope_from(list_t *acc_list, const char *from)
{
    account_t *a = NULL;
    char *acc_from;
    
    while (!list_is_empty(acc_list))
    {
	acc_list = acc_list->next;
	acc_from = ((account_t *)(acc_list->data))->from;
	if (acc_from && strcmp(from, acc_from) == 0)
	{
	    a = acc_list->data;
	    break;
	}
    }

    return a;
}


/*
 * is_on(), is_off()
 *
 * Check whether the given string is "on" or "off"
 */

int is_on(char *s)
{
    return (strcmp(s, "on") == 0);
}

int is_off(char *s)
{
    return (strcmp(s, "off") == 0);
}


/*
 * check_auth_arg()
 *
 * see conf.h
 */

int check_auth_arg(char *arg)
{
    size_t l, i;

    if (*arg == '\0')
    {
	return 0;
    }
    else if (strcmp(arg, "plain") == 0 
	    || strcmp(arg, "cram-md5") == 0
	    || strcmp(arg, "digest-md5") == 0
	    || strcmp(arg, "gssapi") == 0
	    || strcmp(arg, "login") == 0
	    || strcmp(arg, "ntlm") == 0)
    {
    	l = strlen(arg);
	for (i = 0; i < l; i++)
	{
	    arg[i] = toupper((unsigned char)arg[i]);
	}
	return 0;
    }
    else
    {
	return 1;
    }
}


/*
 * check_dsn_notify_arg()
 *
 * see conf.h
 */

int check_dsn_notify_arg(char *arg)
{
    int count;
    size_t i;
    size_t l;

    if (strcmp(arg, "never") != 0)
    {
	l = 0;
	count = 0;
	if (strstr(arg, "failure"))
	{
	    count++;
	    l += 7;
	}
	if (strstr(arg, "delay"))
	{
	    count++;
	    l += 5;
	}
	if (strstr(arg, "success"))
	{
	    count++;
	    l += 7;
	}
	if (count == 0 
		|| (strlen(arg) != l + count - 1)
		|| (count == 2 && !strchr(arg, ','))
		|| (count == 3 && !(strchr(arg, ',') 
			&& strchr(strchr(arg, ',') + 1, ','))))
	{
	    return 1;
	}
    }
    l = strlen(arg);
    for (i = 0; i < l; i++)
    {
	arg[i] = toupper((unsigned char)arg[i]);
    }
    return 0;
}

    
/*
 * override_account()
 *
 * see conf.h
 */

void override_account(account_t *acc1, account_t *acc2)
{
    if (acc2->mask & ACC_HOST)
    {
	free(acc1->host);
	acc1->host = acc2->host ? xstrdup(acc2->host) : NULL;
    }
    if (acc2->mask & ACC_PORT)
    {
	acc1->port = acc2->port;
    }
    if (acc2->mask & ACC_DOMAIN)
    {
	free(acc1->domain);
	acc1->domain = acc2->domain ? xstrdup(acc2->domain) : NULL;
    }
    if (acc2->mask & ACC_FROM)
    {
	free(acc1->from);
	acc1->from = acc2->from ? xstrdup(acc2->from) : NULL;
    }
    if (acc2->mask & ACC_AUTH_MECH)
    {
	free(acc1->auth_mech);
	acc1->auth_mech = acc2->auth_mech ? xstrdup(acc2->auth_mech) : NULL;
    }
    if (acc2->mask & ACC_USERNAME)
    {
	free(acc1->username);
	acc1->username = acc2->username ? xstrdup(acc2->username) : NULL;
    }
    if (acc2->mask & ACC_PASSWORD)
    {
	free(acc1->password);
	acc1->password = acc2->password ? xstrdup(acc2->password) : NULL;
    }
    if (acc2->mask & ACC_NTLMDOMAIN)
    {
	free(acc1->ntlmdomain);
	acc1->ntlmdomain = acc2->ntlmdomain ? xstrdup(acc2->ntlmdomain) : NULL;
    }
    if (acc2->mask & ACC_TLS)
    {
	acc1->tls = acc2->tls;
    }
    if (acc2->mask & ACC_TLS_KEY_FILE)
    {
	free(acc1->tls_key_file);
	acc1->tls_key_file = acc2->tls_key_file ? xstrdup(acc2->tls_key_file) : NULL;
    }
    if (acc2->mask & ACC_TLS_CERT_FILE)
    {
	free(acc1->tls_cert_file);
	acc1->tls_cert_file = acc2->tls_cert_file ? xstrdup(acc2->tls_cert_file) : NULL;
    }
    if (acc2->mask & ACC_TLS_TRUST_FILE)
    {
	free(acc1->tls_trust_file);
	acc1->tls_trust_file = acc2->tls_trust_file ? xstrdup(acc2->tls_trust_file) : NULL;
    }
    if (acc2->mask & ACC_TLS_NOSTARTTLS)
    {
	acc1->tls_nostarttls = acc2->tls_nostarttls;
    }
    if (acc2->mask & ACC_TLS_NOCERTCHECK)
    {
	acc1->tls_nocertcheck = acc2->tls_nocertcheck;
    }
    if (acc2->mask & ACC_DSN_RETURN)
    {
	free(acc1->dsn_return);
	acc1->dsn_return = acc2->dsn_return ? xstrdup(acc2->dsn_return) : NULL;
    }
    if (acc2->mask & ACC_DSN_NOTIFY)
    {
	free(acc1->dsn_notify);
	acc1->dsn_notify = acc2->dsn_notify ? xstrdup(acc2->dsn_notify) : NULL;
    }
    if (acc2->mask & ACC_KEEPBCC)
    {
	acc1->keepbcc = acc2->keepbcc;
    }
    if (acc2->mask & ACC_LOGFILE)
    {
	free(acc1->logfile);
	acc1->logfile = acc2->logfile ? xstrdup(acc2->logfile) : NULL;
    }
    acc1->mask |= acc2->mask;
}


/*
 * set_default_port()
 *
 * If the port in the given account is not set, set it to the default value,
 * which depends on tls and tls_nostarttls.
 */

void set_default_port(account_t *acc)
{
    struct servent *se;
    
    if (!(acc->mask & ACC_PORT))
    {
	if (acc->tls && acc->tls_nostarttls)
	{
	    se = getservbyname("ssmtp", NULL);
	    acc->port = se ? ntohs(se->s_port) : 465;
	}
	else
	{
	    se = getservbyname("smtp", NULL);
	    acc->port = se ? ntohs(se->s_port) : 25;
	}
    }
}


/*
 * check_account()
 * 
 * see conf.h
 */

int check_account(account_t *acc, char *errstr)
{
    if (!acc->host)
    {
	snprintf(errstr, errstr_bufsize, "no host information");
	return CONF_ESYNTAX;
    }
    if (acc->tls_key_file && !acc->tls_cert_file) 
    {
	snprintf(errstr, errstr_bufsize, "TLS key file requires TLS cert file");
	return CONF_ESYNTAX;
    }
    if (!acc->tls_key_file && acc->tls_cert_file)
    {
	snprintf(errstr, errstr_bufsize, "TLS cert file requires TLS key file");
	return CONF_ESYNTAX;
    }
    if (acc->tls_nocertcheck && acc->tls_trust_file)
    {
	snprintf(errstr, errstr_bufsize, "cannot use TLS trust file with TLS cert check turned off");
	return CONF_ESYNTAX;
    }

    return CONF_EOK;
}


/* 
 * some small helper functions
 */

int skip_blanks(char *s, int i)
{
    while (s[i] == ' ' || s[i] == '\t')
    {
	i++;
    }
    return i;
}

int get_length(char *s)
{
    int i = 0;
    
    while (s[i] != '\0' && s[i] != ' ' && s[i] != '\t')
    {
	i++;
    }
    return i;
}

/* get index of last non-blank character. -1 means there is none. */
int get_last_nonblank(char *s)
{
    int i;

    i = (int)strlen(s) - 1;
    while (i >= 0 && (s[i] == ' ' || s[i] == '\t'))
    {
	i--;
    }
    return i;
}


/*
 * get_next_cmd()
 * 
 * Read a line from 'f'. Split it in a command part (first word after
 * whitespace) and an argument part (the word after the command).
 * Whitespace is ignored.
 * Sets the flag 'empty_line' if the line is empty.
 * Sets the flag 'eof' if EOF occured.
 * On errors, 'empty_line' and 'eof', 'cmd' and 'arg' NULL.
 * On success, 'cmd' and 'arg' are allocated strings.
 * Used error codes: CONF_EIO, CONF_EPARSE
 */

int get_next_cmd(FILE *f, char **cmd, char **arg, int *empty_line, int *eof,
	char *errstr)
{
    char line[LINEBUFSIZE];
    char *p;
    int i;
    int l;
    
    *eof = 0;
    *empty_line = 0;
    *cmd = NULL;
    *arg = NULL;
    if (!fgets(line, (int)sizeof(line), f))
    {
	if (ferror(f))
	{
	    snprintf(errstr, errstr_bufsize, "input error");
	    return CONF_EIO;
	}
	else /* EOF */
	{
	    *eof = 1;
	    return CONF_EOK;
	}
    }
    
    /* Kill '\n'. Beware: sometimes the last line of a file has no '\n' */
    if ((p = strchr(line, '\n')))
    {
	*p = '\0';
    }
    else if (strlen(line) == LINEBUFSIZE - 1)
    {
	snprintf(errstr, errstr_bufsize, 
		"line longer than %d characters", LINEBUFSIZE - 1);
	return CONF_EPARSE;
    }

    i = skip_blanks(line, 0);
    
    if (line[i] == '#' || line[i] == '\0')
    {
	*empty_line = 1;
	return CONF_EOK;
    }

    l = get_length(line + i);
    *cmd = xmalloc((l + 1) * sizeof(char));
    strncpy(*cmd, line + i, l);
    (*cmd)[l] = '\0';
    
    i = skip_blanks(line, i + l);
    l = get_last_nonblank(line + i);
    if (l >= 1 && line[i] == '"' && line[i + l] == '"')
    {
	*arg = xmalloc(l * sizeof(char));
	strncpy(*arg, line + i + 1, l - 1);
	(*arg)[l - 1] = '\0';
    }
    else
    {
	*arg = xmalloc((l + 2) * sizeof(char));
	strncpy(*arg, line + i, l + 1);
	(*arg)[l + 1] = '\0';
    }
    
    return CONF_EOK;
}

	
/* 
 * read_conffile()
 * 
 * Read configuration data from 'f' and store it in 'acc_list'.
 * The name of the configuration file, 'conffile', will be stored in the
 * "conffile" field of each account.
 * Unless an error code is returned, 'acc_list' will always be a new list;
 * it may be empty if no accounts were found.
 * Used error codes: CONF_EIO, CONF_EPARSE, CONF_ESYNTAX
 */

int read_conffile(const char *conffile, FILE *f, list_t **acc_list, char *errstr)
{
    int e;
    list_t *p;
    account_t *acc;
    int line;
    char *cmd;
    char *arg;
    int empty_line;
    int eof;


    *acc_list = list_new();
    p = *acc_list;
    acc = NULL;
    e = CONF_EOK;
    
    for (line = 1; ; line++)
    {
	if ((e = get_next_cmd(f, &cmd, &arg, &empty_line, &eof, 
			errstr)) != CONF_EOK)
	{
	    break;
	}
	if (empty_line)
	{
	    continue;
	}
	if (eof)
	{
	    break;
	}

	if (!acc && strcmp(cmd, "account") != 0)
	{
	    acc = account_new(conffile, "default");
	    list_insert(p, acc);
	    p = p->next;
	}
	if (strcmp(cmd, "account") == 0)
	{
	    if (*arg == '\0')
	    {
		snprintf(errstr, errstr_bufsize, 
			"line %d: command %s needs an argument", line, cmd);
		e = CONF_ESYNTAX;
		break;
	    }
	    else if (find_account(*acc_list, arg))
	    {
		snprintf(errstr, errstr_bufsize,
			"line %d: account %s was already defined", line, arg);
		e = CONF_ESYNTAX;
		break;
	    }
	    else
	    {
		acc = account_new(conffile, arg);
		list_insert(p, acc);
		p = p->next;
	    }
	}
	else if (strcmp(cmd, "host") == 0)
	{
	    acc->mask |= ACC_HOST;
	    if (*arg == '\0')
	    {
		snprintf(errstr, errstr_bufsize, 
			"line %d: command %s needs an argument", line, cmd);
		e = CONF_ESYNTAX;
	    }
	    else
	    {
		free(acc->host);
		acc->host = xstrdup(arg);
	    }
	}
	else if (strcmp(cmd, "port") == 0)
	{
	    acc->mask |= ACC_PORT;
	    if (*arg == '\0')
	    {
		snprintf(errstr, errstr_bufsize, 
			"line %d: command %s needs an argument", line, cmd);
		e = CONF_ESYNTAX;
		break;
	    }
	    else
	    {
		acc->port = atoi(arg);
		if (acc->port < 1 || acc->port > 65535)
		{
		    snprintf(errstr, errstr_bufsize,
			    "line %d: invalid port number %d", line, acc->port);
		    e = CONF_ESYNTAX;
		    break;
		}
	    }
	}
	else if (strcmp(cmd, "domain") == 0)
	{
	    acc->mask |= ACC_DOMAIN;
	    if (*arg == '\0')
	    {
		snprintf(errstr, errstr_bufsize, 
			"line %d: command %s needs an argument", line, cmd);
		e = CONF_ESYNTAX;
		break;
	    }
	    else
	    {
		free(acc->domain);
		acc->domain = xstrdup(arg);
	    }
	}
	else if (strcmp(cmd, "from") == 0)
	{
	    acc->mask |= ACC_FROM;
	    if (*arg == '\0')
	    {
		snprintf(errstr, errstr_bufsize, 
			"line %d: command %s needs an argument", line, cmd);
		e = CONF_ESYNTAX;
		break;
	    }
	    else
	    {
		free(acc->from);
		acc->from = xstrdup(arg);
	    }
	}
	else if (strcmp(cmd, "auth") == 0)
	{
	    acc->mask |= ACC_AUTH_MECH;
	    free(acc->auth_mech);
	    if (is_off(arg))
	    {
		acc->auth_mech = NULL;
	    }
	    else if (is_on(arg))
	    {
		acc->auth_mech = xstrdup("");
	    }
	    else if (check_auth_arg(arg) == 0)
	    {
		acc->auth_mech = xstrdup(arg);
	    }
	    else
	    {
		snprintf(errstr, errstr_bufsize, 
			"line %d: invalid argument %s for command %s", line, arg, cmd);
		e = CONF_ESYNTAX;
		break;
	    }
	}
	else if (strcmp(cmd, "user") == 0)
	{
	    acc->mask |= ACC_USERNAME;
	    free(acc->username);
	    if (*arg == '\0')
	    {
		acc->username = NULL;
	    }
	    else
	    {
		acc->username = xstrdup(arg);
	    }
	}
	else if (strcmp(cmd, "password") == 0)
	{
	    acc->mask |= ACC_PASSWORD;
	    free(acc->password);
	    if (*arg == '\0')
	    {
		acc->password = NULL;
	    }
	    else
	    {
		acc->password = xstrdup(arg);
	    }
	}
	else if (strcmp(cmd, "ntlmdomain") == 0)
	{
	    acc->mask |= ACC_NTLMDOMAIN;
	    free(acc->ntlmdomain);
	    if (*arg == '\0')
	    {
		acc->ntlmdomain = NULL;
	    }
	    else
	    {
		acc->ntlmdomain = xstrdup(arg);
	    }
	}
	else if (strcmp(cmd, "tls") == 0)
	{
	    acc->mask |= ACC_TLS;
	    if (*arg == '\0' || is_on(arg))
	    {
		acc->tls = 1;
	    }
	    else if (is_off(arg))
	    {
		acc->tls = 0;
	    }
	    else
	    {
		snprintf(errstr, errstr_bufsize, 
			"line %d: invalid argument %s for command %s", line, arg, cmd);
		e = CONF_ESYNTAX;
		break;
	    }
	}
	else if (strcmp(cmd, "tls_key_file") == 0)
	{
	    acc->mask |= ACC_TLS_KEY_FILE;
	    free(acc->tls_key_file);
	    if (*arg == '\0')
	    {
		acc->tls_key_file = NULL;
	    }
	    else
	    {
		if (expand_tilde(arg, &(acc->tls_key_file), errstr) != PATHS_EOK)
		{
		    e = CONF_ESYNTAX;
		    break;
		}
	    }
	}
	else if (strcmp(cmd, "tls_cert_file") == 0)
	{
	    acc->mask |= ACC_TLS_CERT_FILE;
	    free(acc->tls_cert_file);
	    if (*arg == '\0')
	    {
		acc->tls_cert_file = NULL;
	    }
	    else
	    {
		if (expand_tilde(arg, &(acc->tls_cert_file), errstr) != PATHS_EOK)
		{
		    e = CONF_ESYNTAX;
		    break;
		}
	    }
	}
	else if (strcmp(cmd, "tls_trust_file") == 0)
	{
	    acc->mask |= ACC_TLS_TRUST_FILE;
	    free(acc->tls_trust_file);
	    if (*arg == '\0')
	    {
		acc->tls_trust_file = NULL;
	    }
	    else
	    {
		if (expand_tilde(arg, &(acc->tls_trust_file), errstr) != PATHS_EOK)
		{
		    e = CONF_ESYNTAX;
		    break;
		}
	    }
	}
	else if (strcmp(cmd, "tls_certcheck") == 0)
	{
	    acc->mask |= ACC_TLS_NOCERTCHECK;
	    if (*arg == '\0' || is_on(arg))
	    {
		acc->tls_nocertcheck = 0;
	    }
	    else if (is_off(arg))
	    {
		acc->tls_nocertcheck = 1;
	    }
	    else
	    {
		snprintf(errstr, errstr_bufsize, 
			"line %d: invalid argument %s for command %s", line, arg, cmd);
		e = CONF_ESYNTAX;
		break;
	    }
	}
	else if (strcmp(cmd, "tls_starttls") == 0)
	{
	    acc->mask |= ACC_TLS_NOSTARTTLS;
	    if (*arg == '\0' || is_on(arg))
	    {
		acc->tls_nostarttls = 0;
	    }
	    else if (is_off(arg))
	    {
		acc->tls_nostarttls = 1;
	    }
	    else
	    {
		snprintf(errstr, errstr_bufsize, 
			"line %d: invalid argument %s for command %s", line, arg, cmd);
		e = CONF_ESYNTAX;
		break;
	    }
	}
	else if (strcmp(cmd, "dsn_return") == 0)
	{
	    acc->mask |= ACC_DSN_RETURN;
	    free(acc->dsn_return);
	    if (*arg == '\0')
	    {
		acc->dsn_return = NULL;
	    }
	    else
	    {
		if (strcmp(arg, "headers") == 0)
		{
		    acc->dsn_return = xstrdup("HDRS");
		}
		else if (strcmp(arg, "full") == 0)
		{
		    acc->dsn_return = xstrdup("FULL");
		}
		else
		{
		    snprintf(errstr, errstr_bufsize,
			    "line %d: invalid argument %s for command %s", line, arg, cmd);
		    e = CONF_ESYNTAX;
		    break;
		}
	    }
	}
	else if (strcmp(cmd, "dsn_notify") == 0)
	{
	    acc->mask |= ACC_DSN_NOTIFY;
	    free(acc->dsn_notify);
	    if (*arg == '\0')
	    {
		acc->dsn_notify = NULL;
	    }
	    else
	    {
		if (check_dsn_notify_arg(arg) != 0)
		{
		    snprintf(errstr, errstr_bufsize,
			    "line %d: invalid argument %s for command %s", line, arg, cmd);
		    e = CONF_ESYNTAX;
		}
		acc->dsn_notify = xstrdup(arg);
		break;
	    }
	}
	else if (strcmp(cmd, "keepbcc") == 0)
	{
	    acc->mask |= ACC_KEEPBCC;
	    if (*arg == '\0' || is_on(arg))
	    {
		acc->keepbcc = 1;
	    }
	    else if (is_off(arg))
	    {
		acc->keepbcc = 0;
	    }
	    else
	    {
	   	snprintf(errstr, errstr_bufsize,
    			"line %d: invalid argument %s for command %s", line, arg, cmd);
		e = CONF_ESYNTAX;
		break;
	    }
	}
	else if (strcmp(cmd, "logfile") == 0)
	{
	    acc->mask |= ACC_LOGFILE;
	    free(acc->logfile);
	    if (*arg == '\0')
	    {
		acc->logfile = NULL;
	    }
	    else
	    {
		if (expand_tilde(arg, &(acc->logfile), errstr) != PATHS_EOK)
		{
		    e = CONF_ESYNTAX;
		    break;
		}
	    }
	}
	/* support the deprecated commands tls_nocertcheck and tls_nostarttls, too */
	else if (strcmp(cmd, "tls_nocertcheck") == 0)
	{
	    acc->mask |= ACC_TLS_NOCERTCHECK;
	    if (*arg == '\0')
	    {
		acc->tls_nocertcheck = 1;
	    }
	    else
	    {
		snprintf(errstr, errstr_bufsize, 
			"line %d: command %s does not take an argument", line, cmd);
		e = CONF_ESYNTAX;
		break;
	    }
	}
	else if (strcmp(cmd, "tls_nostarttls") == 0)
	{
	    acc->mask |= ACC_TLS_NOSTARTTLS;
	    if (*arg == '\0')
	    {
		acc->tls_nostarttls = 1;
	    }
	    else
	    {
		snprintf(errstr, errstr_bufsize, 
			"line %d: command %s does not take an argument", line, cmd);
		e = CONF_ESYNTAX;
		break;
	    }
	}
	else
	{
	    snprintf(errstr, errstr_bufsize, 
		    "line %d: unknown command %s", line, cmd);
	    e = CONF_ESYNTAX;
	    break;
	}
	free(cmd);
	free(arg);
    }
    free(cmd);
    free(arg);
    
    if (e != CONF_EOK)
    {
	list_xfree(*acc_list, account_free);
	*acc_list = NULL;
    }

    return e;
}


/*
 * get_conf()
 * 
 * see conf.h
 */

int get_conf(const char *conffile, list_t **acc_list, char *errstr)
{
    FILE *f;
    int e;

    if (!(f = fopen(conffile, "r")))
    {
	snprintf(errstr, errstr_bufsize, "%s", strerror(errno));
	return CONF_EIO;
    }

    if ((e = read_conffile(conffile, f, acc_list, errstr)) != CONF_EOK)
    {
	fclose(f);
	return e;
    }
    fclose(f);
    
    if (list_is_empty(*acc_list))
    {
	list_free(*acc_list);
	snprintf(errstr, errstr_bufsize, "no valid configuration data");
	return CONF_ESYNTAX;
    }

    return CONF_EOK;
}
